
#include <memory.h>
#include <string.h>
#include <iostream.h>
#include <stdlib.h>
#include "bi++.h"

/*
 * if we aren't in a multitasking environment, we can assign static numbers
 * for temporary calculations
 */

#ifndef _MT_
static bigint bitmp1, bitmp2, bitmp3, bitmp4;
#endif

/*
 * constructors and destructors
 */

bigint::bigint ()
{
  length = first = 0;
  sign   = 0;

  memset (digits, 0, BIGLENGTH * sizeof (unsigned short));
}

bigint::bigint (const bigint &other)
{
  memcpy (digits, other.digits, BIGLENGTH * sizeof (unsigned short));

  first  = other.first;
  length = other.length;
  sign   = other.sign;
}

bigint::bigint (const char * const string)
{
  memset (digits, 0, BIGLENGTH * sizeof (unsigned short));
  
  first  = 0;
  length = 0;
  sign   = 0;

  operator= (string);
}

bigint::~bigint ()
{
}

/*
 * Input and Output
 */

ostream &
operator<< (ostream &ostr, const bigint &num)
{
  unsigned short *ptr=num.digits+num.length;
  long i=num.length-num.first;

  ostr.form ("%s%u",
	     (num.sign < 0) ? "-" : "",
	     *ptr--);

  while (i--)
    ostr.form ("%04u", *ptr--);

  if (num.first)
    ostr.form ("e%ld", 4 * num.first);
}

/*
 * assignments
 */

bigint &
bigint::operator= (const bigint &other)
{
  if (&other == this)
    return *this;

  if (length >= other.length) {
    if (first > other.length)
      memset (digits + first, 0,
	      (length - first + 1) * sizeof (unsigned short));
    else {
      memset (digits + other.length + 1, 0,
	      (length - other.length) * sizeof (unsigned short));
      
      if (other.first > first)
	memset (digits + first, 0,
		(other.first - first) * sizeof (unsigned short));
    }
  }
  else if (other.length > length) {
    if (other.first > length)
      memset (digits + first, 0,
	      (length - first + 1) * sizeof (unsigned short));
    else if (other.first > first)
      memset (digits + first, 0,
	      (other.first - first) * sizeof (unsigned short));
  }
  memcpy (digits  + other.first,
	  other.digits + other.first,
	  (other.length - other.first + 1) * sizeof (unsigned short));
  
  length = other.length;
  first  = other.first;
  sign   = other.sign;

  return *this;
}

bigint &
bigint::operator= (const long zahl)
{
  long thezahl, ziffer;
  
  memset (digits + first, 0,
	  (length - first + 1) * sizeof (unsigned short));

  if (zahl > 0) {
    thezahl = zahl;
    sign = 1;
  }
  else if (zahl < 0) {
    thezahl = - zahl;
    sign = -1;
  }
  else if (zahl == 0) {
    sign = length = first = 0L;
    return *this;
  }
  length = 0L;
  first  = 0L;

  if ((ziffer = thezahl % 10000) != 0) {
    *digits = (unsigned) ziffer;
  }
  if ((ziffer = (thezahl % 100000000) / 10000) != 0) {
    *(digits + 1) = (unsigned) ziffer;
    length = 1L;
    if (*digits == 0)
      first = 1L;
  }
  if ((ziffer = thezahl / 100000000) != 0) {
    *(digits + 2) = (unsigned) ziffer;
    length = 2L;
    if (*digits == 0 && *(digits + 1) == 0)
      this->first = 2L;
  }
  return *this;
}

bigint &
bigint::operator= (const char * const string)
{
  const char *ptr1 = string, *ptr2;
  long slen, texp;

  memset (digits + first, 0,
	  (length - first + 1) * sizeof (unsigned short));

  if (ptr1 == NULL) {
    first  = 0;
    length = 0;
    sign   = 0;
    return *this;
  }
  while (*ptr1 == ' ' || *ptr1 == '\t')
    ptr1++;
  if (*ptr1 == '+') {
    ptr1++;
    sign =  1;
  }
  else if (*ptr1 == '-') {
    ptr1++;
    sign = -1;
  }
  else
    sign =  1;

  if ((ptr2 = strpbrk (string, "Ee")) != NULL) {
    slen = ptr2 - ptr1;
    texp = atoi (ptr2+1);
  }
  else {
    slen = strlen (ptr1);
    texp = 0;
  }
  first  = texp;
  length = texp + (slen-1) / 4;

  switch (slen % 4) {
  case 1:
    digits[texp + slen/4] = *ptr1 - '0';
    slen -= 1; ptr1 += 1;
    break;
  case 2:
    digits[texp + slen/4] = 10 * *ptr1 + *(ptr1 + 1) - (11 * '0');
    slen -= 2; ptr1 += 2;
    break;
  case 3:
    digits[texp + slen/4] = 
      100 * *ptr1 + 10 * *(ptr1 + 1) + *(ptr1 + 2) - (111 * '0');
    slen -= 3; ptr1 += 3;
    break;
  }
  while (slen) {
    slen -= 4;
    digits[texp + slen/4] =
      1000 * *ptr1 + 100 * *(ptr1 + 1) + 10 * *(ptr1 + 2) + *(ptr1 + 3) -
	(1111 * '0');
    ptr1 += 4;
  }
  return adjust ();
}

/*
 * adjust length and first values of a number. needed frequently
 */

bigint &
bigint::adjust (void)
{
  unsigned short *ptr = digits + length;

  while (*ptr-- == 0 && length > 0)
    length--;

  if (length != 0) {
    ptr = digits + first;

    while (*ptr++ == 0)
      first++;
  }
  else
    first = 0;

  if (first == 0 && length == 0 && *digits == 0)
    sign = 0;

  return *this;
}

/*
 * add numbers without looking at their signs. subfunction of
 * operator+=
 */

bigint &
bigint::add_unsigned (const bigint &other)
{
  unsigned short *ptr1, *ptr2, temp;
  long alength, afirst;

  if (other.sign == 0L)
    return *this;

  /*
   * if they don't overlap, simply copy.
   */
  if (other.length < first) {
    memcpy (digits  + other.first,
	    other.digits + other.first,
	    (other.length - other.first + 1) * sizeof (unsigned short));
    first = other.first;
    return *this;
  }
  else if (other.first > length) {
    memcpy (digits  + other.first,
	    other.digits + other.first,
	    (other.length - other.first + 1) * sizeof (unsigned short));
    alength = other.length;
    return *this;
  }
  /*
   * they do overlap at least partially. Copy the parts that don't.
   */
  if (first > other.first) {
    memcpy (digits  + other.first,
	    other.digits + other.first,
	    (first - other.first) * sizeof (unsigned short));
    afirst = first;
  }
  else
    afirst = other.first;

  if (length < other.length) {
    memcpy (digits  + length + 1,
	    other.digits + length + 1,
	    (other.length - length) * sizeof (unsigned short));
    alength = length - afirst;
  }
  else
    alength = other.length - afirst;

  /*
   * add overlapping section
   */
  ptr1 = digits  + afirst;
  ptr2 = other.digits + afirst;

 kein_uebertrag:
  if (alength < 0)
    goto ende;

  temp = *ptr1 + *ptr2++;
  alength--;
  
  if (temp < 10000) {
    *ptr1++ = temp;
    goto kein_uebertrag;
  }

 mit_uebertrag:
  *ptr1++ = temp - 10000;

  if (alength < 0)
    goto ende_mit_uebertrag;

  temp = *ptr1 + *ptr2++ + 1;
  alength--;

  if (temp < 10000) {
    *ptr1++ = temp;
    goto kein_uebertrag;
  }
  goto mit_uebertrag;

 ende_mit_uebertrag:
  *ptr1 += 1;
  if (*ptr1 == 10000) {
      *ptr1++ = 0;
      goto ende_mit_uebertrag;
  }

 ende:
  if (length > other.length)
    length += 1;
  else
    length  = other.length + 1;

  if (first > other.first)
    first = other.first;

  return adjust ();
}  

/*
 * substract without looking at the sign. The function cannot handle zero
 * wraparounds, it can only substract a lesser number from a bigger number.
 * Thus, *this must be bigger than other.
 */

bigint &
bigint::sub_unsigned (const bigint &other)
{
  unsigned short *ptr1, *ptr2;
  long alength, temp, afirst, flag;

  if (other.sign == 0L)
    return *this;

  /*
   * if other does not even reach into this, then decrement first digit of
   * this, copy 9999-other into first and fill space inbetween with 9999
   */
  if (other.length < first) {
    ptr1 = digits + other.first;
    ptr2 = other.digits + other.first;
    alength = other.first + 1;

    *ptr1++ = 10000 - *ptr2++;

    while (alength <= other.length) {
      *ptr1++ = 9999 - *ptr2++;
      alength++;
    }
    while (alength < first) {
      *ptr1++ = 9999;
      alength++;
    }
    /*
     * we're pointing at first
     */
    if (*ptr1 == 0) {
      while (*ptr1 == 0)
	*ptr1++ = 9999;
    }
    *ptr1 -= 1;
    first  = other.first;

    return adjust ();
  }
  /*
   * they do overlap at least parially.
   */
  ptr1 = digits  + other.first;
  ptr2 = other.digits + other.first;

  if (first > other.first) {
    /*
     * same procedure as above
     */
    alength = other.first + 1;
    flag    = 1;

    *ptr1++ = 10000 - *ptr2++;

    while (alength <= first) {
      *ptr1++ = 9999 - *ptr2++;
      alength++;
    }
  }
  else {
    alength = other.first;
    flag    = 0;
  }

  while (alength <= other.length) {
    temp = *ptr1 - *ptr2++ - flag;

    if (temp >= 0) {
      *ptr1++ = (unsigned) temp;
      flag    = 0;
    }
    else {
      *ptr1++ = (unsigned) (10000 + temp);
      flag    = 1;
    }
    alength++;
  }
  if (flag) {
    while (*ptr1 == 0)
      *ptr1++ = 9999;
    *ptr1 -= 1;
  }
  if (first > other.first)
    first = other.first;

  return adjust ();
}

int
bigint::cmp_unsigned (const bigint &other) const
{
  unsigned short *ptr1, *ptr2;
  long alength;

  if (length > other.length)
    return 1;
  else if (other.length > length)
    return -1;

  ptr1    = digits       + length;
  ptr2    = other.digits + other.length;
  alength = length  - ((first>other.first)?first:other.first) + 1;

  while (*ptr1 == *ptr2 && alength>0) {
    ptr1--; ptr2--; alength--;
  }
  if (*ptr1 > *ptr2)
    return 1;
  else if (*ptr1 < *ptr2)
    return -1;

  if (first < other.first)
    return 1;
  else if (first > other.first)
    return -1;

  return 0;
}

bigint &
bigint::addi_unsigned (const long number)
{
  long ziffer, ubt=0, index;

  ziffer = number % 10000;

  if (ziffer != 0) {
    digits[0] += ziffer;
    first      = 0;

    if (digits[0] >= 10000) {
      digits[0] -= 10000;
      ubt        = 1;
    }
  }

  ziffer = (number % 100000000) / 10000 + ubt;
  ubt    = 0;

  if (ziffer != 0) {
    digits[1] += ziffer;

    if (digits[1] >= 10000) {
      digits[1] -= 10000;
      ubt        = 1;
    }
    if (first > 1)
      first = 1;
    if (length < 1)
      length = 1;
  }
  ziffer = number / 100000000;

  if (ziffer != 0) {
    digits[2] += ziffer;

    if (digits[2] >= 10000) {
      digits[2] -= 10000;
      ubt        = 1;
    }
    if (first > 2)
      first = 2;
    if (length < 2)
      length = 2;
  }

  if (ubt) {
    index = 3;
    digits[index] += 1;

    while (digits[index] > 10000) {
      digits[index]   -= 10000;
      digits[++index] += 1;
    }
    if (index > length)
      length = index;
  }

  return *this;
}

/*
 * operators
 */

bigint &
bigint::operator+= (const bigint &other)
{
  int flag;

  if (sign == other.sign)
    return add_unsigned (other);

  if (sign == 0L)
    return (*this = other);
  else if (other.sign == 0L)
    return *this;

  /*
   * now we know their sign is different. check out which one's bigger
   */

  if ((flag = cmp_unsigned (other)) == 0)
    operator= (0);
  else if (flag > 0)
    sub_unsigned (other);
  else {
#ifdef _MT_
    bigint bitmp1 (other);
#else
    bitmp1 = other;
#endif

    bitmp1.sub_unsigned (*this);
    *this = bitmp1;
  }

  return *this;
}

bigint &
bigint::operator-= (const bigint &other)
{
  /*
   * why complicated if it can be done simply?
   * usually, we'd twist the sign of other, but it's defined as const, and
   * shall stay so. But, a = a - b = - (-a + b)
   */

  sign = - sign;
  operator+= (other);
  sign = - sign;

  return *this;
}

/*
 * shift function is needed later. multiplies number with 10^digits
 * num.shiftleft(2) = num * 100 but much faster
 */

bigint &
bigint::shiftleft (const long numdig)
{
  unsigned short *ptr, on=0, nn;
  long shift = numdig / 4, i;

  if (shift != 0) {
    memmove (digits + first + shift,
	     digits + first,
	     (length - first + 1) * sizeof (unsigned short));
    memset  (digits + first, 0, shift * sizeof (unsigned short));

    length += shift;
    first  += shift;
  }
  if ((numdig % 4) == 0)
    return *this;

  length += 1;
  ptr     = digits + first;

  switch (numdig % 4) {
  case 1:
    for (i=length-first; i>=0; i--) {
      nn     = *ptr / 1000;
      *ptr++ = (*ptr % 1000) * 10 + on;
      on     = nn;
    }
    break;
  case 2:
    for (i=length-first; i>=0; i--) {
      nn     = *ptr / 100;
      *ptr++ = (*ptr % 100) * 100 + on;
      on     = nn;
    }
    break;
  case 3:
    for (i=length-first; i>=0; i--) {
      nn     = *ptr / 10;
      *ptr++ = (*ptr % 10) * 1000 + on;
      on     = nn;
    }
    break;
  } /* switch */

  return adjust ();
}

bigint &
bigint::mult_nnnn (const long number)
{
  long uebertrag = 0L, ergebnis;
  unsigned short *ptr1 = digits;
  long i, ziffer;

  if (number > 0)
    ziffer = number;
  else if (number < 0) {
    ziffer = - number;
    sign   = - sign;
  }
  else if (number == 0) {
    return (*this = 0);
  }
  if (sign == 0)
    return *this;

  ptr1 = digits + first;
  i    = length - first + 2;
  
  for (; i>=0; i--) {
    ergebnis  = (long) *ptr1 * ziffer + uebertrag;
    *ptr1++   = (unsigned short) (ergebnis % 10000);
    uebertrag = ergebnis / 10000;
  }
  length++;

  return adjust ();
}

bigint
bigint::operator* (const bigint &other) const
{
  const unsigned short *ptr2, *ptr3;
  long uebertrag=0L, ergebnis, i, j;
  unsigned short *ptr1;
  bigint result;

  if (sign == 0 || other.sign == 0)
    return result;

  result.sign = sign * other.sign;

  ptr2 = digits + first;
  for (i=first; i<=length; i++, ptr2++) {
    ptr3      = other.digits  + other.first;
    ptr1      = result.digits + other.first + i;
    j         = other.first;
    uebertrag = 0L;

    while (j<=other.length || uebertrag) {
      ergebnis  = (long) *ptr2 * (long) *ptr3 + uebertrag + *ptr1;
      *ptr1++   = (unsigned short) (ergebnis % 10000);
      uebertrag = ergebnis / 10000;
      j++; ptr3++;
    }
  }

  /*
   * calculate new length. Play verrry safe here
   */

  result.length = (i+1) + (j+1) + 1;
  result.first  = first + other.first;

  /*
   * adjust length and first values
   */

  return result.adjust ();
}

bigint
bigint::operator* (const long zahl) const
{
  long ziffer, operand;
  bigint result;
  
  if (sign == 0)
    return result;

  if (zahl > 0) {
    result.sign = sign;
    operand     = zahl;
  }
  else if (zahl < 0) {
    result.sign = - sign;
    operand     = - zahl;
  }
  else if (zahl == 0)
    return result;

  {
#ifdef _MT_
    bigint bitmp1;
#else
    bitmp1 = 0;
#endif
    
    ziffer = operand % 10000L;

    if (ziffer == operand) {
      result = *this;
      return result.mult_nnnn (ziffer);
    }

    if (ziffer > 0) {
      result = *this;
      if (ziffer > 1)
	result.mult_nnnn (ziffer);
    }
    ziffer = (operand % 100000000L) / 10000L;

    if (ziffer > 0) {
      bitmp1 = *this;
      if (ziffer > 1)
	bitmp1.mult_nnnn (ziffer);
      result += bitmp1.shiftleft (4);
    }

    ziffer = operand / 100000000;

    if (ziffer > 0) {
      bitmp1 = *this;
      if (ziffer > 1)
	bitmp1.mult_nnnn (ziffer);
      result += bitmp1.shiftleft (8);
    }
  }
  return result;
}

bigint &
bigint::operator*= (const long zahl)
{
  long ziffer, operand;
  
  if (sign == 0)
    return *this;

  if (zahl > 0)
    operand = zahl;
  else if (zahl < 0) {
    operand = - zahl;
    sign    = - sign;
  }
  else if (zahl == 0)
    return (*this = 0);

  {
#ifdef _MT_
    bigint bitmp1, bitmp2;
#else
    bitmp1 = 0;
    bitmp2  = 0;
#endif
    
    ziffer = operand % 10000L;

    if (ziffer == operand)
      return mult_nnnn (ziffer);

    if (ziffer > 0) {
      bitmp2 = *this;
      if (ziffer > 1)
	bitmp2.mult_nnnn (ziffer);
    }
    ziffer = (operand % 100000000L) / 10000L;

    if (ziffer > 0) {
      bitmp1 = *this;
      if (ziffer > 1)
	bitmp1.mult_nnnn (ziffer);
      bitmp2 += bitmp1.shiftleft (4);
    }

    ziffer = operand / 100000000;

    if (ziffer > 0) {
      bitmp1 = *this;
      if (ziffer > 1)
	bitmp1.mult_nnnn (ziffer);
      bitmp2 += bitmp1.shiftleft (8);
    }
    *this = bitmp2;
  }
  return *this;
}

bigint
bigint::operator/ (const bigint &other) const
{
  long sig1=sign, sig2=other.sign, factor, ldiff;
  bigint result;
  int cres;

  if (sign == 0)
    return result;
  else if (other.sign == 0)
    return result; /* division by zero ignored */

  if ((cres = cmp_unsigned (other)) < 0) {
    return result;
  }
  else if (cres == 0) {
    /*
     * numbers are equal
     */
    return (result = 1);
  }

  {
    /*
     * assign temporary numbers
     */
#ifdef _MT_
    bigint bitmp2, bitmp3, bitmp4(*this);
#else
    bitmp4 = *this;
#endif

    result.sign = sign * other.sign;
    bitmp4.sign = 1;
    
    while (bitmp4.sign) {
      if (bitmp4.length < other.length)
	break;
      
      factor = *(bitmp4.digits+bitmp4.length) / *(other.digits+other.length);
      ldiff  = bitmp4.length - other.length;

      if (factor == 0) {
	if (bitmp4.length == bitmp4.first || bitmp4.length == other.length)
	  break;
	else {
	  factor = (10000 * *(bitmp4.digits + bitmp4.length) +
		    *(bitmp4.digits + bitmp4.length - 1)) /
		      *(other.digits + other.length);
	  ldiff--;
	}
      }
      bitmp2  = (factor * bitmp4.sign);
      bitmp2.shiftleft (4*ldiff);
      bitmp3  = bitmp2 * other;
      bitmp4 -= bitmp3;
      result += bitmp2;
    }
    if (bitmp4.sign == -1) {
      bitmp2 = 1;
      result.sub_unsigned (bitmp2);
    }
  }

  return result;
}

int
bigint::operator< (const bigint &other) const
{
  if (sign < other.sign)
    return 1;
  else if (sign > other.sign)
    return 0;
  else
    return (cmp_unsigned(other) < 0) ? 1 : 0;
}

int
bigint::operator> (const bigint &other) const
{
  if (sign > other.sign)
    return 1;
  else if (sign < other.sign)
    return 0;
  else
    return (cmp_unsigned(other) > 0) ? 1 : 0;

}

int
bigint::operator== (const bigint &other) const
{
  if (sign != other.sign)
    return 0;
  else
    return (cmp_unsigned(other) == 0) ? 1 : 0;
}

int
bigint::operator<= (const bigint &other) const
{
  if (sign < other.sign)
    return 1;
  else if (sign > other.sign)
    return 0;
  else
    return (cmp_unsigned(other) <= 0) ? 1 : 0;
}

int
bigint::operator>= (const bigint &other) const
{
  if (sign > other.sign)
    return 1;
  else if (sign < other.sign)
    return 0;
  else
    return (cmp_unsigned(other) >= 0) ? 1 : 0;
}

int
bigint::operator== (const long zahl) const
{
  long ziffer, uzahl;

  if (zahl < 0) {
    if (sign >= 0)
      return 0;
    uzahl = - zahl;
  }
  else if (zahl == 0)
    return (sign == 0) ? 1 : 0;
  else {
    if (sign <= 0)
      return 0;
    uzahl = zahl;
  }
  if (length > 2)
    return 0;

  ziffer = uzahl % 10000;
  if (digits[0] != ziffer)
    return 0;

  ziffer = (uzahl % 100000000) / 10000;
  if (digits[1] != ziffer)
    return 0;

  ziffer = uzahl / 100000000;
  if (digits[2] != ziffer)
    return 0;

  return 1;
}

bigint
operator+ (const long number, const bigint &other)
{
#ifdef _MT_
  bigint bitmp1;
#endif
  bitmp1 = number;
  return (bitmp1 + other);
}

bigint
operator- (const long number, const bigint &other)
{
#ifdef _MT_
  bigint bitmp1;
#endif
  bitmp1 = number;
  return (bitmp1 - other);
}

bigint &
bigint::operator+= (const long number)
{
#ifdef _MT_
  bigint bitmp4;
#endif
  bitmp4 = number;

  return (*this += bitmp4);
}

bigint &
bigint::operator-= (const long number)
{
#ifdef _MT_
  bigint bitmp4;
#endif
  bitmp4 = number;

  return (*this -= bitmp4);
}

bigint
bigint::operator% (const bigint &other) const
{
  return (*this - (*this / other) * other);
}

/*
 * unary operations
 */

bigint &
bigint::operator++ (void)
{
  unsigned short *ptr = digits;

  if (sign < 0) {
    if (first == 0 && length == 0 && *ptr == 1) {
      *ptr = 0;
      sign = 0;
    }
    else {
      first = 0;
      while (*ptr == 0)
	*ptr++ = 9999;
      *ptr -= 1;
    }
  }
  else if (sign == 0) {
    *ptr = 1;
    sign = 1;
  }
  else /* if (sign > 0) */ {
    first = 0;
    while (*ptr == 9999)
      *ptr++ = 0;
    *ptr   += 1;
    length += 1;
  }
  return adjust ();
}

bigint &
bigint::operator-- (void)
{
  unsigned short *ptr = digits;

  if (sign > 0) {
    if (first == 0 && length == 0 && *ptr == 1) {
      *ptr = 0;
      sign = 0;
    }
    else {
      first = 0;
      while (*ptr == 0)
	*ptr++ = 9999;
      *ptr -= 1;
    }
  }
  else if (sign == 0) {
    *ptr =  1;
    sign = -1;
  }
  else /* if (sign < 0) */ {
    first = 0;
    while (*ptr == 9999)
      *ptr++ = 0;
    *ptr   += 1;
    length += 1;
  }
  return adjust ();
}

bigint &
bigint::operator- (void)
{
  sign = - sign;
}
