#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include "bigint.h"

static bigint ztemp, ziel, dtemp;
static int zinit=1;

void bi_const (bigint *this)
{
  this->length  = 0L;
  this->first   = 0L;
  this->sign    = 1L;

  /*
   * if ((this->digits = ptr = malloc (BIGLENGTH * sizeof (short))) == NULL)
   *   exit (1);
   * for (i=0L; i<BIGLENGTH; i++)
   *   *ptr++ = 0;
   */

  memset (this->digits, 0, BIGLENGTH * sizeof (unsigned short));
  *(this->digits) = 1;
}

void bi_one (bigint *this)
{
  memset (this->digits + this->first, 0,
	  (this->length - this->first + 1) * sizeof (unsigned short));
  this->length     = 0L;
  this->first      = 0L;
  this->sign       = 1L;
  *(this->digits)  = 1;
}

void bi_null (bigint *this)
{
  memset (this->digits + this->first, 0,
	  (this->length - this->first + 1) * sizeof (unsigned short));
  this->length = 0L;
  this->first  = 0L;
  this->sign   = 0L;
}

void bi_dest (bigint *this)
{
  /*
   * if (this->digits != NULL)
   *   free (this->digits);
   */
  this->length  = 0L;
  this->first   = 0L;
  /*
   * this->digits = NULL;
   */
}

bigint *bi_assign_long  (bigint *this, long zahl)
{
  long thezahl, ziffer;

  memset (this->digits + this->first, 0,
	  (this->length - this->first + 1) * sizeof (unsigned short));

  if (zahl > 0) {
    thezahl = zahl;
    this->sign = 1;
  }
  else if (zahl < 0) {
    thezahl = - zahl;
    this->sign = -1;
  }
  else if (zahl == 0) {
    this->sign = this->length = this->first = 0L;
    return this;
  }
  this->length = 0L;
  this->first  = 0L;

  if ((ziffer = thezahl % 10000) != 0) {
    *(this->digits) = (unsigned) ziffer;
  }
  if ((ziffer = (thezahl % 100000000) / 10000) != 0) {
    *(this->digits + 1) = (unsigned) ziffer;
    this->length = 1L;
    if (*(this->digits) == 0)
      this->first = 1L;
  }
  if ((ziffer = thezahl / 100000000) != 0) {
    *(this->digits + 2) = (unsigned) ziffer;
    this->length = 2L;
    if (*(this->digits) == 0 && *(this->digits + 1) == 0)
      this->first = 2L;
  }
  return this;
}

bigint *bi_assign (bigint *this, bigint *other)
{
  if (this->length >= other->length) {
      if (this->first > other->length)
	memset (this->digits + this->first, 0,
		(this->length - this->first + 1) * sizeof (unsigned short));
      else {
	  memset (this->digits + other->length + 1, 0,
		  (this->length - other->length) * sizeof (unsigned short));
	  
	  if (other->first > this->first)
	    memset (this->digits + this->first, 0,
		    (other->first - this->first) * sizeof (unsigned short));
      }
  }
  else if (other->length > this->length) {
      if (other->first > this->length)
	memset (this->digits + this->first, 0,
		(this->length - this->first + 1) * sizeof (unsigned short));
      else if (other->first > this->first)
	memset (this->digits + this->first, 0,
		(other->first - this->first) * sizeof (unsigned short));
  }
  memcpy (this->digits  + other->first,
	  other->digits + other->first,
	  (other->length - other->first + 1) * sizeof (unsigned short));
  
  this->length = other->length;
  this->first  = other->first;
  this->sign   = other->sign;
  return this;
}

/*
 * comparison functions
 */

int bi_compare (bigint *this, bigint *other)
{
  unsigned short *ptr1, *ptr2;
  long length;

  if (this->sign > other->sign)
    return 1;
  else if (this->sign < other->sign)
    return -1;
  else if (this->sign == 0L && other->sign == 0L)
    return 0;

  if (this->length > other->length)
    return (this->sign>0) ? 1 : -1;
  else if (other->length > this->length)
    return (this->sign>0) ? -1 : 1;

  ptr1   = this->digits  + this->length;
  ptr2   = other->digits + other->length;
  length = this->length  - ((this->first>other->first)?this->first:other->first) + 1;

  while (*ptr1 == *ptr2 && length>0) {
    ptr1--; ptr2--; length--;
  }
  if (*ptr1 > *ptr2)
    return (this->sign>0) ? 1 : -1;
  else if (*ptr1 < *ptr2)
    return (this->sign>0) ? -1 : 1;

  if (this->first < other->first)
    return (this->sign>0) ? 1 : -1;
  else if (this->first > other->first)
    return (this->sign>0) ? -1 : 1;

  return 0;
}

/*
 * compare without looking at the sign. needed for adding and substracting
 */

static int bi_unsigned_compare (bigint *this, bigint *other)
{
  unsigned short *ptr1, *ptr2;
  long length;

  if (this->length > other->length)
    return 1;
  else if (other->length > this->length)
    return -1;

  ptr1   = this->digits  + this->length;
  ptr2   = other->digits + other->length;
  length = this->length  - ((this->first>other->first)?this->first:other->first) + 1;

  while (*ptr1 == *ptr2 && length>0) {
    ptr1--; ptr2--; length--;
  }
  if (*ptr1 > *ptr2)
    return 1;
  else if (*ptr1 < *ptr2)
    return -1;

  if (this->first < other->first)
    return 1;
  else if (this->first > other->first)
    return -1;

  return 0;
}

/*
 * add with equal sign (to positive numbers or two negative numbers)
 */

static bigint *bi_add_eqs (bigint *this, bigint *other)
{
  register unsigned short *ptr1, *ptr2, temp;
  register long length;
  long afirst;

  if (other->sign == 0L)
    return this;

  /*
   * if they don't overlap, simply copy.
   */
  if (other->length < this->first) {
    memcpy (this->digits  + other->first,
	    other->digits + other->first,
	    (other->length - other->first + 1) * sizeof (unsigned short));
    this->first = other->first;
    return this;
  }
  else if (other->first > this->length) {
    memcpy (this->digits  + other->first,
	    other->digits + other->first,
	    (other->length - other->first + 1) * sizeof (unsigned short));
    this->length = other->length;
    return this;
  }
  /*
   * they do overlap at least partially. Copy the parts that don't.
   */
  if (this->first > other->first) {
    memcpy (this->digits  + other->first,
	    other->digits + other->first,
	    (this->first - other->first) * sizeof (unsigned short));
    afirst = this->first;
  }
  else
    afirst = other->first;

  if (this->length < other->length) {
    memcpy (this->digits  + this->length + 1,
	    other->digits + this->length + 1,
	    (other->length - this->length) * sizeof (unsigned short));
    length = this->length - afirst;
  }
  else
    length = other->length - afirst;

  /*
   * add overlapping section
   */
  ptr1 = this->digits  + afirst;
  ptr2 = other->digits + afirst;

 kein_uebertrag:
  if (length < 0)
    goto ende;

  temp = *ptr1 + *ptr2++;
  length--;
  
  if (temp < 10000) {
    *ptr1++ = temp;
    goto kein_uebertrag;
  }

 mit_uebertrag:
  *ptr1++ = temp - 10000;

  if (length < 0)
    goto ende_mit_uebertrag;

  temp = *ptr1 + *ptr2++ + 1;
  length--;

  if (temp < 10000) {
    *ptr1++ = temp;
    goto kein_uebertrag;
  }
  goto mit_uebertrag;

 ende_mit_uebertrag:
  *ptr1 += 1;
  if (*ptr1 == 10000) {
      *ptr1++ = 0;
      goto ende_mit_uebertrag;
  }

 ende:
  if (this->length > other->length)
    this->length += 1;
  else
    this->length  = other->length + 1;

  if (this->first > other->first)
    this->first   = other->first;

  ptr1 = this->digits + this->length;

  while (*ptr1-- == 0 && this->length > 0)
    this->length -= 1;

  if (this->length != 0) {
    ptr1 = this->digits + this->first;
    while (*ptr1++ == 0)
      this->first += 1;
  }
  else
    this->first = 0;
  
  return this;
}  

/*
 * substract (add with nonequal sign) the lesser number from the bigger one
 */

static bigint *bi_subst_neqs (bigint *this, bigint *other)
{
  register unsigned short *ptr1, *ptr2;
  register long length, temp;
  long afirst, flag;

  if (other->sign == 0L)
    return this;

  /*
   * if other does not even reach into this, then decrement first digit of
   * this, copy 9999-other into first and fill space inbetween with 9999
   */
  if (other->length < this->first) {
    ptr1 = this->digits + other->first;
    ptr2 = other->digits + other->first;
    length = other->first + 1;

    *ptr1++ = 10000 - *ptr2++;

    while (length <= other->length) {
      *ptr1++ = 9999 - *ptr2++;
      length++;
    }
    while (length < this->first) {
      *ptr1++ = 9999;
      length++;
    }
    /*
     * we're pointing at this->first
     */
    if (*ptr1 == 0) {
      while (*ptr1 == 0)
	*ptr1++ = 9999;
    }
    *ptr1 -= 1;

    this->first = other->first;

    ptr1 = this->digits + this->length;
    while (*ptr1-- == 0 && this->length > 0)
      this->length -= 1;

    if (this->length != 0) {
      ptr1 = this->digits + this->first;
      while (*ptr1++ == 0)
	this->first += 1;
    }
    else
      this->first = 0;
    
    return this;
  }
  /*
   * they do overlap at least parially.
   */
  ptr1 = this->digits  + other->first;
  ptr2 = other->digits + other->first;

  if (this->first > other->first) {
    /*
     * same procedure as above
     */
    length = other->first + 1;
    flag   = 1;

    *ptr1++ = 10000 - *ptr2++;

    while (length <= this->first) {
      *ptr1++ = 9999 - *ptr2++;
      length++;
    }
  }
  else {
    length = other->first;
    flag   = 0;
  }

  while (length <= other->length) {
    temp = *ptr1 - *ptr2++ - flag;

    if (temp >= 0) {
      *ptr1++ = (unsigned) temp;
      flag    = 0;
    }
    else {
      *ptr1++ = (unsigned) (10000 + temp);
      flag    = 1;
    }
    length++;
  }
  if (flag) {
    while (*ptr1 == 0)
      *ptr1++ = 9999;
    *ptr1 -= 1;
  }
  if (this->first > other->first)
    this->first = other->first;

  ptr1 = this->digits + this->length;

  while (*ptr1-- == 0 && this->length > 0)
    this->length -= 1;

  if (this->length != 0) {
    ptr1 = this->digits + this->first;
    while (*ptr1++ == 0)
      this->first += 1;
  }
  else
    this->first = 0;
  
  return this;
}

/*
 * add and substract with respect to sign
 */

bigint *bi_add (bigint *this, bigint *other)
{
  int flag;

  if (this->sign == other->sign)
    return bi_add_eqs (this, other);

  if (this->sign == 0L)
    return bi_assign (this, other);
  else if (other->sign == 0L)
    return this;

  flag = bi_unsigned_compare (this, other);

  if (flag == 0) {
    bi_null (this);
  }
  else if (flag > 0)
    bi_subst_neqs (this, other);
  else {
    /*
     * because bi_subst_neqs needs the bigger number as first parameter
     */
    if (zinit) {
      bi_const (&ztemp);
      bi_const (&ziel);
      bi_const (&dtemp);
      zinit = 0;
    }
    bi_assign     (&ztemp, other);
    bi_subst_neqs (&ztemp, this);
    bi_assign     (this,  &ztemp);
  }
  return this;
}

bigint *bi_subst (bigint *this, bigint *other)
{
  int flag;

  other->sign = - other->sign;
  bi_add (this, other);
  other->sign = - other->sign;

  return this;
}

bigint *bi_mult_nnnn (bigint *this, long zahl)
{
  register long uebertrag = 0L, ergebnis;
  register unsigned short *ptr1 = this->digits;
  long i, ziffer;

  if (zahl > 0)
    ziffer = zahl;
  else if (zahl < 0) {
    ziffer = -zahl;
    this->sign = - this->sign;
  }
  else if (zahl == 0) {
    bi_null (this);
    return this;
  }
  if (this->sign == 0L)
    return this;

  ptr1 = this->digits + this->first;
  i    = this->length  - this->first + 2;
  
  for (; i>=0; i--) {
    ergebnis  = (long) *ptr1 * ziffer + uebertrag;
    *ptr1++   = (unsigned short) (ergebnis % 10000);
    uebertrag = ergebnis / 10000;
  }
  this->length += 3;
  ptr1          = this->digits + this->length;

  while (*ptr1-- == 0 && this->length > 0)
    this->length -= 1;

  if (this->length != 0) {
    ptr1 = this->digits + this->first;
    while (*ptr1++ == 0)
      this->first += 1;
  }
  else
    this->first = 0;

  return this;
}

bigint *bi_multiply (bigint *dest, bigint *src1, bigint *src2)
{
  register long uebertrag = 0L, ergebnis;
  register unsigned short *ptr1, *ptr2, *ptr3;
  long i,j;

  bi_null (dest);

  if (src1->sign == 0L || src2->sign == 0L)
    return dest;

  dest->sign = src1->sign * src2->sign;

  ptr2 = src1->digits + src1->first;
  for (i=src1->first; i<=src1->length; i++, ptr2++) {
    ptr3      = src2->digits + src2->first;
    ptr1      = dest->digits + src2->first + i;
    j         = src2->first;
    uebertrag = 0L;

    while (j<=src2->length || uebertrag) {
      ergebnis  = (long) *ptr2 * (long) *ptr3 + uebertrag + *ptr1;
      *ptr1++   = (unsigned short) (ergebnis % 10000);
      uebertrag = ergebnis / 10000;
      j++; ptr3++;
    }
  }

  dest->length = (i+1) + (j+1) + 1;
  dest->first  = src1->first + src2->first;
  ptr1         = dest->digits + dest->length;

  while (*ptr1-- == 0 && dest->length > 0)
    dest->length -= 1;

  if (dest->length != 0) {
    ptr1 = dest->digits + dest->first;
    while (*ptr1++ == 0)
      dest->first += 1;
  }
  else
    dest->first = 0;

  return dest;
}

bigint *bi_divide (bigint *dest, bigint *src1, bigint *src2)
{
  long sig1=src1->sign, sig2=src2->sign, factor, ldiff;
  int cres;

  if (zinit) {
    bi_const (&ztemp);
    bi_const (&ziel);
    bi_const (&dtemp);
    zinit = 0;
  }

  if (src1->sign == 0L) {
    bi_null (dest);
    return dest;
  }
  else if (src2->sign == 0L) {
    bi_null (dest); /* division by zero ignored ... */
    return dest;
  }
  if ((cres = bi_unsigned_compare (src1, src2)) < 0) {
    bi_null (dest);
    return dest;
  }
  else if (cres == 0) {
    bi_one (dest);
    return dest;
  }

  bi_assign (&ztemp, src1);
  bi_null   (dest);
  dest->sign = src1->sign * src2->sign;

  while (ztemp.sign) {
    if (ztemp.length < src2->length)
      break;

    factor = *(ztemp.digits+ztemp.length) / *(src2->digits+src2->length);
    ldiff  = ztemp.length - src2->length;

    if (factor == 0) {
      if (ztemp.length == ztemp.first || ztemp.length == src2->length)
	break;
      else {
	factor = (10000 * *(ztemp.digits + ztemp.length) +
		  *(ztemp.digits + ztemp.length - 1)) /
		    *(src2->digits+src2->length);
	ldiff--;
      }
    }
    bi_assign_long (&ziel, factor);
    bi_shift_left  (&ziel, 4 * ldiff);
    bi_multiply    (&dtemp, &ziel, src2);
    bi_subst_neqs  (&ztemp, &dtemp);
    bi_add_eqs     (dest,  &ziel);
  }

  return dest;
}

bigint *bi_shift_left (bigint *this, long stellen)
{
  long shift = stellen / 4, i;
  register unsigned short *ptr, on=0, nn;
  
  if (shift != 0) {
    memmove (this->digits + this->first + shift,
	     this->digits + this->first,
	     (this->length - this->first + 1) * sizeof (unsigned short));
    memset  (this->digits + this->first, 0, shift * sizeof (unsigned short));

    this->length += shift;
    this->first  += shift;
  }
  if ((stellen % 4) == 0)
    return this;
  
  this->length += 1;
  ptr           = this->digits + this->first;
    
  switch (stellen % 4) {
  case 1:
    for (i=this->length-this->first; i>=0; i--) {
      nn     = *ptr / 1000;
      *ptr++ = (*ptr % 1000) * 10 + on;
      on     = nn;
    }
    break;
  case 2:
    for (i=this->length-this->first; i>=0; i--) {
      nn     = *ptr / 100;
      *ptr++ = (*ptr % 100) * 100 + on;
      on     = nn;
    }
    break;
  case 3:
    for (i=this->length-this->first; i>=0; i--) {
      nn     = *ptr / 10;
      *ptr++ = (*ptr % 10) * 1000 + on;
      on     = nn;
    }
    break;
  } /* switch */
  
  ptr = this->digits + this->length;
  
  while (*ptr-- == 0 && this->length > 0)
    this->length -= 1;
  
  if (this->length != 0) {
    ptr = this->digits + this->first;
    while (*ptr++ == 0)
      this->first += 1;
  }
  else
    this->first = 0;

  return this;
}

bigint *bi_mult_long (this, zahl)
     register bigint *this;
     register long zahl;
{
  register long ziffer;
  long operand;
  
  if (this->sign == 0L)
    return this;
  if (zahl > 0)
    operand = zahl;
  else if (zahl < 0) {
    operand = - zahl;
    this->sign = - this->sign;
  }
  else if (zahl == 0) {
    bi_null (this);
    return this;
  }

  ziffer = operand % 10000L;
  
  if (zinit) {
      bi_const (&ztemp);
      bi_const (&ziel);
      bi_const (&dtemp);
      zinit = 0;
  }
  else {
      bi_one   (&ztemp);
      bi_one   (&ziel);
  }

  if (ziffer == operand) {
      bi_mult_nnnn (this, ziffer);
      return this;
  }
  if (ziffer > 0L) {
    bi_assign (&ziel, this);

    if (ziffer > 1L)
      bi_mult_nnnn (&ziel, ziffer);
  }
  else
    bi_null (&ziel);
  
  ziffer = (operand % 100000000L) / 10000L;
  if (ziffer > 0L) {
    bi_assign (&ztemp, this);

    if (ziffer > 1L)
      bi_mult_nnnn (&ztemp, ziffer);

    bi_add (&ziel, bi_shift_left (&ztemp, 4L));
  }
  
  ziffer = operand / 100000000L;
  if (ziffer > 0L) {
    bi_assign (&ztemp, this);

    if (ziffer > 1L)
      bi_mult_nnnn (&ztemp, ziffer);

    bi_add (&ziel, bi_shift_left (&ztemp, 8L));
  }
  bi_assign (this, &ziel);

  return this;
}

void bi_output (bigint *this, FILE *outfile)
{
  unsigned short *ptr = this->digits + this->length;
  long gedruckt = -1L, i=this->length;

  for (; i>=0; i--) {
    if (++gedruckt % 15L == 0)
      fprintf (outfile, "\n %6ld : %c", 4L*gedruckt,
	       (i==this->length)?((this->sign<0)?'-':' '):' ');
    
    fprintf (outfile, "%04u", *ptr--);
  }
  fprintf (outfile, "\n");
}

void bi_output_plain (bigint *this, FILE *outfile)
{
  unsigned short *ptr = this->digits+this->length;
  long i=this->length-this->first;

  fprintf (outfile, "%s%u",
	   (this->sign<0)?"-":"",
	   *ptr--);

  while (i--)
    fprintf (outfile, "%04u", *ptr--);

  if (this->first)
    fprintf (outfile, "e%ld", 4*this->first);
}

/*
void main (int argc, char *argv[])
{
  bigint fakultaet;
  long ende = atol (argv[1]), i;
  time_t startzeit, endzeit;

  bi_const (&fakultaet);
  fprintf  (stderr, "Calculating the faculty of %ld (w) '93-95 /Frank\n\n", ende);
	
  startzeit = time (NULL);
  
  for (i=2; i<=ende; i++) {
    if (i%10 == 0)
      fprintf (stderr, "%ld. Iteration\x0d", i);

    bi_mult_long (&fakultaet, i);
  }
  
  endzeit   = time (NULL);
  
  printf ("\nFaculty of %ld \n(w) Frank Pilhofer\n", ende);
  bi_output (&fakultaet);
  printf ("\nCalculation time in secs: %ld\n", endzeit - startzeit);
}
*/
